/*
 * MainBoard_UART.c
*************************************************************************
	Author: Alexander von Lhmann 2014
	
	Hardware
   	processor:		AtMega644A
	clock:			20MHz external RC oscillator

	Function			Pin		Word	Connected With


	Comm. Module 0
	TX_RST_M0			PD7				PD4 (RXRST_M1)
	TXRDY_M0			PB2				PD3 (RXR_M1)
	TX2_M0				PB1				PD2 (RX2_M1)
	TX1_M0				PB3				PD1 (RX1_M1)

						PA0				ADC0 / Photosensor-Trigger
	Comm. Module 1
	TX_RST_M1			PD5				PD4 (RXRST_M2)
	TXRDY_M1			PA3				PD3 (RXR_M2)
	TX2_M1				PA4				PD2 (RX2_M2)
	TX1_M1				PA2				PD1 (RX1_M2)

	Comm. Module 2
	TX_RST_M2			PD4				PD4 (RXRST_M0)
	TXRDY_M2			PC7				PD3 (RXR_M0)
	TX2_M2				PA1				PD2 (RX2_M0)
	TX1_M2				PB0				PD1 (RX1_M0)

	Comm. Module 3
	TX_RST_M3			PD6				PD4 (RXRST_M3)
	TXRDY_M3			PA6				PD3 (RXR_M3)
	TX2_M3				PA7				PD2 (RX2_M3)
	TX1_M3				PA5				PD1 (RX1_M3)

	UART:
	RXD0				PD0				TX_BT
	TXD0				PD1				RX_BT
	RTS					PD2				RTS#BT
	CTS					PD3				CTS#BT
	
	SPI:
	MOSI				PB5				SDI ADC
	MISO				PB6				SDO ADC
	SS					PB4				+5V
	SCK					PB7				SCK ADC
	Slave Select		PC0				CS ADC
	
	Bluetooth Reset		PC3				BT_RST
	
	LED1				PC4				
	LED2				PC6			
	
	
	L0 := 750 nm (bright red LED)
	L1 := 850 nm (hard to see red LED)
	

*************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include "stdbool.h"
#include <string.h>
#include <stdio.h>

#define F_CPU 20E6 //System clock in Hz
#define UART_BAUD_RATE 9600
#define ADC_TRIGGER_LEVEL 143	// ~2,8V

#include <util/delay.h>

#include "SPI_functions.h"
#include "uart.h"
#include "LTC2486_functions.h"
#include "Mainboard_UART_functions.h"
#include "sbopcodes_custom.h"



// Global VARIABLES
uint16_t timer_ms_counter;								// x*10ms-Time-Value	


// Timer Compare Match A Interrupt (each ~10ms)
ISR(TIMER1_COMPA_vect) {
	timer_ms_counter++;
}


int main(void) {
	
	/* Variables */
	volatile uint16_t adc_config =0;														// ADC-Config Data (for LTC2468)
	volatile uint16_t adc_value =0;															// Extracted ADC-Value
	volatile uint32_t adc_receive =0;														// Received ADC Value
	volatile uint16_t temp_value =0;														// Received and extracted Temp-Value
	uint8_t adc =0;																			// ADC-Trigger Value
	unsigned int c;
	bool adc_toggle =1;
	conf_stat config = {true, true, false, false, false, false, false, {{1,0,0,0},{1,0,0,0},{1,0,0,0},{1,0,0,0}},{'0','1','2','3','X'},0, 0, false, 0};		// Config/Status of Machine, initialized with default M0 all Channels active
	meas_info info = {0,0,false,false,false};
			
	/*********************** Initialization ***********************/
	//Port Directions
	
	DDRA= 0b11111110;				// NIRS-Module 1-3 Communication + Sensor Threshold ADC Trigger								
	DDRB= 0b10101111;				// NIRS-Module 3-4 Communication, SPI										
	DDRC= 0b11011001;				// SPI Slave select, LEDs										
	DDRD= 0b11111010;				// UART
	
	PORTC |= (1<<PINC3);			// BT_Reset auf High;
	
	//Initializations
	uart_init(UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU));					// UART initialization
	SPI_init();						// SPI initialization
	Timer_init();					// Timer initialization	
	ADC_init();						// ADC initialization
	sei();							// Global Interrupt enable
	Bluetooth_Reset();				// Reset Bluetooth Module
		
	uint16_t adcbuffer =0;
		
    while(1) {	
		
        c = uart_getc();				// Get received character from ringbuffer: lower byte = received character, higher byte (bitmask) = last receive error
        if ( c & UART_NO_DATA ){			// no data available from UART 
			if (config.running_flag && !config.temp_flag){	/* Running Mode: Configure ADC and send data */	
				
				// 10BIT ADC Version
				channel_administrator(&config, &info);							// Calculate next active Channel and buffer last configuration in "info"
				NIRS_module_set_channel(config.active_mod, config.active_ch);	// Set next active channel (NIRS Module)
				
				NIRS_module_trigger(config.active_mod);							// Activate NIRS Module
				_delay_ms(15);													// dwell time
				adc_value = ADC_gettwobytevalue();
				_delay_ms(1);							
				NIRS_module_deactivate(config.active_mod);						// Deactivate NIRS Module trigger line
				//NIRS_module_reset(config.active_mod);							// turn off channel
				_delay_ms(15);
				
				if (config.empty_buf_flag==false){								// If Transmission is not started anew: Send data, else don't send (empties ADC-Buffer with old Value)
					send_ADC_data(info.active_mod, info.active_ch, info.lambda, info.speedmode, adcbuffer, timer_ms_counter);	// Send Data
				}
				config.empty_buf_flag = false;
				
				adcbuffer = adc_value;
				
				// SPI 16Bit ADC Version
				/*adc = ADC_get_value();
				ADC_handler(adc, ADC_TRIGGER_LEVEL, &adc_toggle);
				
				channel_administrator(&config, &info);							// Calculate next active Channel and buffer last configuration in "info"
				
				adc_config=generate_configuration(&config);						// Prepare next ADC-Configuration-Word
				NIRS_module_set_channel(config.active_mod, config.active_ch);	// Set next active channel (NIRS Module)
				
				SPI_activate_slave();											// Activate ADC (Read)
				while(!LTC2486_conversion_complete());							// Wait until ADC Value (last setup) is ready;
				NIRS_module_deactivate(config.active_mod);						// Deactivate NIRS Module **
				_delay_ms(1);
				//if (!config.empty_buf_flag){
				NIRS_module_trigger(config.active_mod);							// Activate NIRS Module
				//}
				_delay_ms(4);													// 4ms dwell time
				adc_receive = LTC2486_SPI_get_value(adc_config);				// Get ADC Value, Config for next Measurement
				SPI_deactivate_slave();											// Deactivate ADC/ Start new AD-Conversion									
				
				
				if (config.empty_buf_flag==false){								// If Transmission is not started anew: Send data, else don't send (empties ADC-Buffer with old Value)
					//send_ADC_raw(adc_receive);
					//adc_value = LTC2486_extract_value(adc_receive);				// Extract data
					send_ADC_data(info.active_mod, info.active_ch, info.lambda, info.speedmode, adc_value, timer_ms_counter);	// Send Data
				}
				config.empty_buf_flag = false;
				*/
			}	
			else if (config.temp_flag) {					/* Get Temperature of ADC and send it */
				SPI_activate_slave();
				adc_config=generate_configuration(&config);
				
				while(!LTC2486_conversion_complete());	
				adc_receive = LTC2486_SPI_get_value(adc_config);	// Write Temp-Request in ADC (disregard old value)
				while(!LTC2486_conversion_complete());
				adc_receive = LTC2486_SPI_get_value(adc_config);	// Get Temp-Value (and write another Temp-Request)
				
				temp_value = LTC2486_extract_temp(adc_receive);		// Extract data
				send_ADC_temp(temp_value);							// Send Data
				
				SPI_deactivate_slave();
				config.temp_flag=false;
				config.start_flag = true;			// Start again from first active channel
				config.empty_buf_flag = true;		// Set Buffer_Empty flag: Next transmission will be sent twice
			}
			else {									//Running Mode off: Start next Run with first active channel */
				config.start_flag = true;
			}
        }
        else{
            /*
             * check for Frame or Overrun error
             */
            if ( c & UART_FRAME_ERROR ){							// Framing Error detected, i.e no stop bit detected 
                uart_puts_P("UART Frame Error: ");
            }
            if ( c & UART_OVERRUN_ERROR ){							// Overrun, a character already present in the UART UDR register was not read by the interrupt handler before the next character arrived, one or more received characters have been dropped
                uart_puts_P("UART Overrun Error: ");
            }
            if ( c & UART_BUFFER_OVERFLOW ){						// We are not reading the receive buffer fast enough, one or more received character have been dropped 
                uart_puts_P("Buffer overflow error: ");
            }
			
            process_rxbuf(c, &timer_ms_counter, &config);							// Process Input Buffer				
        }
    }
}